{******************************************************************************}
{                                  ListViewEx
{                                  ----------
{
{                       Copyright  2002 Pieter Zijlstra
{
{ A ListView component in reportstyle which is capable of drawing the little
{ sortdirection arrows (ala Outlook Express).
{
{ E-mail: p.zylstra@hccnet.nl
{ Website: http://home.hccnet.nl/p.zylstra/
{===============================================================================
{                           This software is FREEWARE
{                           -------------------------
{
{ You may freely use it in any software, including commercial software, but
{ be-aware that the code is provided as-is, with no implied warranty.
{===============================================================================
{
{ Notes:
{ - Compared to the original TListView two new properties are added:
{   - property SortOrder   => soNone, soAscending or soDescending
{   - property SortColumn  => 0 based column index.
{ - Original property SortType of TListView is no longer used.
{ - By default the component sorts the columns based on Text (Caption).
{   If you want to sort on Data or any other special sorting
{   (for instance a Date) you'll have to provide an OnCompare
{   event handler.
{   Within this handler you can use the SortOrder and SortColumn
{   properties to find out what and in which direction the items
{   should be sorted.
{ - Delphi version(s) 5, 6 & 7
{
{ Version history:
{ 1.00 02 jul 2002 - First public release.
{ 1.01 26 jul 2002 - Updated comments.
{                  - inherited ColClick is now called at the end instead of
{                    the begin of the procedure. This is done for Listviews.
{                    with OwnerData set to True. The user can now check
{                    the SortColumn and SortOrder in the OnColumnClick
{                    event handler and do his/her own sorting on the
{                    virtual data based on that info.
{                  - Added WM_NCDESTROY handling in HeaderWndProc().
{ 1.10 22 oct 2002 - Handling WM_SIZE message. This is necessary because the
{                    header is wiped out and the arrows needs to be redrawn
{ 1.20 26 dec 2002 - Turned off the deprecated warning for D6 and above.
{ 1.21 03 feb 2003 - Added e-mail and website in comments.
{******************************************************************************}
unit ListViewEx;

interface

{$I CompVers.inc}

// In D6 and up MakeObjectInstance and FreeObjectInstance are moved
// from Forms to Classes. In the Forms unit it is indicated to be
// deprecated. For the moment this warning is turned off.
{$IFDEF DELPHI6_UP}
  {$WARN SYMBOL_DEPRECATED OFF}
{$ENDIF}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Forms, CommCtrl, ComCtrls;

type
  TSortOption = (soNone, soAscending, soDescending);

  TListViewEx = class(TListView)
  private
    FDefHeaderProc: Pointer;
    FHeaderHandle: HWND;
    FHeaderInstance: Pointer;
    FSortOrder: TSortOption;
    FSortColumn: Integer;
    FUpArrow, FDownArrow: TBitmap;
    procedure HeaderWndProc(var Message: TMessage);
    procedure SetColumnBitmap(Index: Integer; Order: TSortOption);
    procedure SetSortColumn(const Value: Integer);
    procedure SetSortOrder(const Value: TSortOption);
    procedure WMParentNotify(var Message: TWMParentNotify); message WM_PARENTNOTIFY;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
  protected
    procedure ColClick(Column: TListColumn); override;
  published
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property SortOrder: TSortOption read FSortOrder write SetSortOrder;
    property SortColumn: Integer read FSortColumn write SetSortColumn;
  end;


procedure Register;

implementation

{$R LISTVIEWARROWS.RES}

procedure Register;
begin
  RegisterComponents('PZGoodies', [TListViewEx]);
end;

function DefaultListViewSort(Item1, Item2: TListItem;
  lParam: Integer): Integer; stdcall;
var
  ColumnIndex: Integer;
  ListView: TListViewEx;
begin
  ListView := TListViewEx(Item1.ListView);
  if Assigned(ListView.OnCompare) then
    ListView.OnCompare(ListView, Item1, Item2, lParam, Result)
  else
  begin
    ColumnIndex := ListView.SortColumn;
    case ColumnIndex of
      0: Result := CompareText(Item1.Caption, Item2.Caption);
    else Result := CompareText(Item1.SubItems[ColumnIndex-1],
                               Item2.SubItems[ColumnIndex-1]);
    end;
    if ListView.SortOrder = soDescending then
      Result := -Result;
  end;
end;

// I don't know a better way at the moment to get a 'transparent'
// bitmap from a Resource other then using a extra bitmap and the
// BrushCopy method.
function CreateBitmapFromResource(ResourceName: string): TBitmap;
var
  bmp: TBitmap;
  rec: TRect;
begin
  Result := TBitmap.Create;
  try
    rec := Rect(0, 0, 8, 8);
    bmp := TBitmap.Create;
    try
      bmp.LoadFromResourceName(HInstance, ResourceName);
      Result.Width  := bmp.Width;
      Result.Height := bmp.Height;
      with Result.Canvas do
      begin
        Brush.Color := clBtnface;
        Brush.Style := bsSolid;
        FillRect(Cliprect);
        BrushCopy(rec, bmp, rec, bmp.TransparentColor);
      end;
    finally
      bmp.Free;
    end;
  except
    Result.Free;
    raise
  end;
end;

constructor TListViewEx.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FUpArrow := CreateBitmapFromResource('LV_ARROWUP');
  FDownArrow := CreateBitmapFromResource('LV_ARROWDOWN');
  ViewStyle := vsReport;
  FHeaderInstance := MakeObjectInstance(HeaderWndProc);
end;

destructor TListViewEx.Destroy;
begin
  FreeObjectInstance(FHeaderInstance);
  if FHeaderHandle <> 0 then
    SetWindowLong(FHeaderHandle, GWL_WNDPROC, LongInt(FDefHeaderProc));
  FDownArrow.Free;
  FUpArrow.Free;
  inherited Destroy;
end;

procedure TListViewEx.ColClick(Column: TListColumn);
begin
  if (FSortOrder = soAscending) and (FSortColumn = Column.Index) then
    FSortOrder := soDescending
  else
    FSortOrder := soAscending;
  SetSortColumn(Column.Index);
  inherited;
end;

procedure TListViewEx.HeaderWndProc(var Message: TMessage);
begin
  try
    with Message do
    begin
      Result := CallWindowProc(FDefHeaderProc, FHeaderHandle, Msg, WParam, LParam);
      case Msg of
        WM_CAPTURECHANGED:
          SetColumnBitmap(SortColumn, SortOrder);

        WM_NCDESTROY:
          begin
            FHeaderHandle := 0;
            FDefHeaderProc := nil;
          end;
      end;
    end;
  except
    Application.HandleException(Self);
  end;
end;

procedure TListViewEx.SetColumnBitmap(Index: Integer; Order: TSortOption);
var
  hdr: HWND;
  Item: THDItem;
begin
  hdr := Listview_GetHeader(Handle);
  FillChar(Item, sizeof(Item), 0);
  Item.Mask := HDI_FORMAT;
  Header_GetItem(hdr, Index, Item);
  if (Order <> soNone) and (Index = FSortColumn) then
  begin
    Item.Mask := Item.Mask or HDI_BITMAP;
    Item.fmt  := Item.fmt  or HDF_BITMAP or HDF_BITMAP_ON_RIGHT;
    if Order = soAscending then
      Item.hbm := FUpArrow.Handle
    else
      Item.hbm := FDownArrow.Handle;
  end
  else
    Item.fmt := Item.fmt and not(HDF_BITMAP or HDF_BITMAP_ON_RIGHT);
  Header_SetItem(hdr, Index, Item);
end;

procedure TListViewEx.SetSortColumn(const Value: Integer);
var
  I: Integer;
begin
  FSortColumn := Value;
  if (FSortColumn >= 0) and (FSortColumn < Columns.Count) then
    if SortOrder <> soNone then
      CustomSort(@DefaultListViewSort, 0);
  for I := 0 to Columns.Count - 1 do
    SetColumnBitmap(I, SortOrder);
end;

procedure TListViewEx.SetSortOrder(const Value: TSortOption);
begin
  if FSortOrder <> Value then
  begin
    FSortOrder := Value;
    SetSortColumn(FSortColumn);
  end;
end;

procedure TListViewEx.WMParentNotify(var Message: TWMParentNotify);
begin
  inherited;
  with Message do
    if (Event = WM_CREATE) and (FHeaderHandle = 0) then
    begin
      FHeaderHandle := ChildWnd;
      FDefHeaderProc := Pointer(GetWindowLong(FHeaderHandle, GWL_WNDPROC));
      SetWindowLong(FHeaderHandle, GWL_WNDPROC, LongInt(FHeaderInstance));
    end;
end;

procedure TListViewEx.WMSize(var Message: TWMSize);
begin
  inherited;
  SetColumnBitmap(SortColumn, SortOrder);
end;

end.
